"""Download required remote files."""

import codecs
import collections
import json
import os
import zipfile

import requests


MTGJSON_ADDRESS = 'http://mtgjson.com/json/'
VERSION_FILENAME = 'version-full.json'
ALLSETS_FILENAME = 'AllSets.json.zip'


class Error(Exception):
    """Base error class for this module."""


class DownloadError(Error):
    """Raised if the downloader fails to fetch a file."""


def fetch_mtgjson(data_path):
    """Check version and fetch (if needed) mtgjson file to data_path."""
    if not os.path.exists(data_path):
        os.makedirs(data_path)
    elif not os.path.isdir(data_path):
        raise DownloadError('data_path: %s must be a folder' % data_path)

    version_filename = os.path.join(data_path, VERSION_FILENAME)
    if os.path.exists(version_filename):
        with open(version_filename, 'r') as version_file:
            local_version_data = json.load(version_file)
            local_version = tuple(
                int(v) for v in local_version_data['version'].split('.'))
    else:
        local_version = (0, 0, 0)

    ver_req = requests.get(MTGJSON_ADDRESS + VERSION_FILENAME)
    if ver_req.status_code != 200:
        raise DownloadError(
            'Could not fetch version: {}'.format(ver_req.reason))
    remote_version_data = ver_req.json()
    remote_version = tuple(
        int(v) for v in remote_version_data['version'].split('.'))

    if local_version >= remote_version:
        return False

    print('Downloading mtgjson data.')
    allsets_filename = os.path.join(data_path, ALLSETS_FILENAME)
    mtg_req = requests.get(MTGJSON_ADDRESS + ALLSETS_FILENAME)
    with open(allsets_filename, 'wb') as allsets_file:
        allsets_file.write(mtg_req.content)
    with open(version_filename, 'wb') as version_file:
        version_file.write(ver_req.content)
    return True


def read_mtgjson(data_path):
    """Read data from mtgjson file and return loaded contents."""
    allsets_filename = os.path.join(data_path, ALLSETS_FILENAME)
    with zipfile.ZipFile(allsets_filename, 'r') as allsets_zipfile:
        [datafilename] = allsets_zipfile.namelist()
        datafile = allsets_zipfile.open(datafilename)
        reader = codecs.getreader('utf-8')
        mtgdata = json.load(
            reader(datafile), object_pairs_hook=collections.OrderedDict)
    return mtgdata



"""CSV Serializer for Deckbox import/export."""

import csv

from mtg_ssm.mtg import models
from mtg_ssm.serialization import interface


DECKBOX_EDITION_TO_RANGE_AND_SETNAMES = {
    'Champs': [(None, None, 'Champs and States')],
    'Commander 2013': [(None, None, 'Commander 2013 Edition')],
    'Commander': [(None, None, 'Magic: The Gathering-Commander')],
    'Conspiracy': [(None, None, 'Magic: The Gathering—Conspiracy')],
    'From the Vault: Annihilation': [
        (None, None, 'From the Vault: Annihilation (2014)')],
    'Introductory Two-Player Set': [(None, None, 'Rivals Quick Start Set')],
    'Magic 2015 Clash Pack Promos': [(None, None, 'Clash Pack')],
    'Magic Game Day Cards': [(None, None, 'Magic Game Day')],
    'Multiverse Gift Box Cards': [(None, None, 'Multiverse Gift Box')],
    'Planechase 2012': [(None, None, 'Planechase 2012 Edition')],
    'Portal Demogame': [(None, None, 'Portal Demo Game')],
    'WotC Online Store': [(None, None, 'Wizards of the Coast Online Store')],
    'WPN/Gateway': [
        (1, 20, 'Gateway'),
        (21, 100, 'Wizards Play Network'),
    ],
}
MTGJSON_TO_DECKBOX_SETNAME = {
    s: k
    for k, v in DECKBOX_EDITION_TO_RANGE_AND_SETNAMES.items()
    for _, _, s in v
}

DECKBOX_HEADER = [
    'Count',
    'Tradelist Count',
    'Name',
    'Edition',
    'Card Number',
    'Condition',
    'Language',
    'Foil',
    'Signed',
    'Artist Proof',
    'Altered Art',
    'Misprint',
    'Promo',
    'Textless',
    'My Price',
]


def get_deckbox_name(card):
    """Given a card, return an appropriate name for deckbox, or None."""
    if card.name != card.names[0]:
        return None

    if card.layout == 'split':
        cardname = ' // '.join(card.names)
    else:
        cardname = card.name

    cardname = cardname.replace('Æ', 'Ae')
    return cardname


def rows_from_printing(printing):
    """Given a CardPrinting yield rows for copies and foils, if present."""
    name = get_deckbox_name(printing.card)
    set_name = printing.set.name
    edition = MTGJSON_TO_DECKBOX_SETNAME.get(set_name, set_name)
    row_base = {
        'Tradelist Count': 0,
        'Edition': edition,
        'Card Number': printing.set_integer,
        'Condition': 'Near Mint',
        'Language': 'English',
        'Signed': None,
        'Artist Proof': None,
        'Altered Art': None,
        'Misprint': None,
        'Promo': 'promo' if printing.set.type_ == 'promo' else None,
        'Textless': None,
        'My Price': None,
    }
    if name is not None:
        row_base['Name'] = name
        copies = printing.counts.get(models.CountTypes.copies, 0)
        foils = printing.counts.get(models.CountTypes.foils, 0)
        if copies:
            # yield {**row_base, 'Foil': None, 'Count': copies}
            copies_row = row_base.copy()
            copies_row.update({'Foil': None, 'Count': copies})
            yield copies_row
        if foils:
            # yield {**row_base, 'Foil': 'foil', 'Count': foils}
            foils_row = row_base.copy()
            foils_row.update({'Foil': 'foil', 'Count': foils})
            yield foils_row


def deckbox_rows_from_collection(coll):
    """Generator that yields csv rows from a collection."""
    for card_set in coll.card_sets:
        for printing in card_set.printings:
            yield from rows_from_printing(printing)


def get_mtgj_setname(edition, number):
    """Use the remappings to get the setname from a deckbox edition."""
    if edition not in DECKBOX_EDITION_TO_RANGE_AND_SETNAMES:
        return edition

    for start, end, setname in DECKBOX_EDITION_TO_RANGE_AND_SETNAMES[edition]:
        if ((start is None or int(number) >= start) and
                (end is None or int(number) <= end)):
            return setname


def create_counts_row(coll, deckbox_row):
    """Given a row from a deckbox csv file, return a counts row."""
    edition = deckbox_row['Edition']
    number = deckbox_row['Card Number']
    mtgj_setname = get_mtgj_setname(edition, number)
    set_code = coll.setname_to_card_set[mtgj_setname].code
    counts = int(deckbox_row['Count']) + int(deckbox_row['Tradelist Count'])
    countname = 'foils' if deckbox_row['Foil'] == 'foil' else 'copies'
    return {
        'name': deckbox_row['Name'].split('//')[0].strip(),
        'set': set_code,
        'number': number,
        countname: counts,
    }


class MtgDeckboxSerializer(interface.MtgSsmSerializer):
    """MtgSsmSerializer for reading/writing deckbox compatible csv files."""

    format = 'deckbox'
    extension = None

    def write_to_file(self, filename: str) -> None:
        """Write the collection to a deckbox csv file."""
        with open(filename, 'w') as csv_file:
            writer = csv.DictWriter(csv_file, DECKBOX_HEADER)
            writer.writeheader()
            for row in deckbox_rows_from_collection(self.collection):
                writer.writerow(row)

    def read_from_file(self, filename: str) -> None:
        """Read collection counts from deckbox csv file."""
        with open(filename, 'r') as deckbox_file:
            reader = csv.DictReader(deckbox_file)
            for row in reader:
                self.load_counts(
                    create_counts_row(self.collection, row),
                    strict=False)



"""Profiling code."""

import cProfile
import io
import pstats
import contextlib


def start():
    """Start and return profiler."""
    profiler = cProfile.Profile()
    profiler.enable()
    return profiler


def finish(profiler):
    """Stop the profiler and print out stats."""
    profiler.disable()
    out_stream = io.StringIO()
    profile_stats = pstats.Stats(
        profiler, stream=out_stream).sort_stats('cumulative')
    profile_stats.print_stats(30)
    print(out_stream.getvalue())


@contextlib.contextmanager
def profiled(enabled):
    """Context manager to profile within a given context."""
    profiler = None
    if enabled:
        profiler = start()
    try:
        yield
    finally:
        if profiler is not None:
            finish(profiler)



"""SQLAlchemy models for managing data."""

import datetime as dt
import enum
import string

VARIANT_CHARS = (string.ascii_letters + '★')


class Card:
    """Model for storing card information."""
    __slots__ = ('collection', 'name', 'layout', 'names')

    def __init__(self, collection, card_data):
        self.collection = collection
        self.name = card_data['name']
        self.layout = card_data['layout']
        self.names = card_data.get('names', [self.name])

    @property
    def strict_basic(self):
        """Is this card one of the five basic lands (not Snow or Wastes)."""
        return self.name in ('Plains', 'Island', 'Swamp', 'Mountain', 'Forest')

    @property
    def printings(self):
        """List of all printings of this card."""
        return self.collection.card_name_to_printings[self.name]

    def __str__(self):
        return 'Card: {card.name}'.format(card=self)

    def __repr__(self):
        return '<Card: {card.name}>'.format(card=self)


class CountTypes(enum.Enum):
    """Enum for possible card printing types (normal, foil)."""
    copies = 'copies'
    foils = 'foils'


class CardPrinting:
    """Model for storing information about card printings."""
    __slots__ = ('collection', 'id_', 'card_name', 'set_code', 'set_number',
                 'set_integer', 'set_variant', 'multiverseid', 'artist',
                 'counts')

    def __init__(self, collection, set_code, card_data):
        self.collection = collection
        self.id_ = card_data['id']
        self.card_name = card_data['name']
        self.set_code = set_code
        self.set_number = card_data.get('number')
        self.multiverseid = card_data.get('multiverseid')
        self.artist = card_data['artist']
        self.counts = {}

        if self.set_number is None:
            self.set_integer = None
            self.set_variant = None
        else:
            self.set_integer = int(self.set_number.strip(VARIANT_CHARS))
            self.set_variant = self.set_number.strip(string.digits) or None

    @property
    def card(self):
        """The Card associated with this printing."""
        return self.collection.name_to_card[self.card_name]

    @property
    def set(self):
        """The CardSet associated with this printing."""
        return self.collection.code_to_card_set[self.set_code]

    def __str__(self):
        return 'CardPrinting: {print.id_}'.format(print=self)

    def __repr__(self):
        return '<CardPrinting: {print.id_}>'.format(print=self)


class CardSet:
    """Model for storing card set information."""
    __slots__ = ('collection', 'code', 'name', 'block', 'release_date',
                 'type_', 'online_only')

    def __init__(self, collection, set_data):
        self.collection = collection
        self.code = set_data['code']
        self.name = set_data['name']
        self.block = set_data.get('block')
        self.release_date = dt.datetime.strptime(
            set_data['releaseDate'], '%Y-%m-%d').date()
        self.type_ = set_data['type']
        self.online_only = set_data.get('onlineOnly', False)

    @property
    def printings(self):
        """The printings in this set.

        Note:
            Printings are ordered by set_integer, set_variant, multiverseid,
            card_name.
        """
        return self.collection.set_code_to_printings[self.code]

    def __str__(self):
        return 'CardSet: {cset.name}'.format(cset=self)

    def __repr__(self):
        return '<CardSet: {cset.code}>'.format(cset=self)



"""Container for holding models."""

import collections

from mtg_ssm.mtg import models


def set_code_to_printings_key(printing):
    """Sort key function for set_code_to_printings index."""
    return (
        printing.set_integer or 0,
        str(printing.set_variant),
        printing.multiverseid or 0,
        printing.card_name,
    )


def card_name_to_printing_key(printing):
    """Sort key function for card_name_to_printings index."""
    return (
        printing.set_code,
        printing.set_integer or 0,
        str(printing.set_variant),
        printing.multiverseid or 0,
    )


class Collection:
    """Container/manager object for storing models."""

    def __init__(self, mtg_json_data=None, include_online_only=False):
        self.name_to_card = {}
        self.code_to_card_set = {}
        self.setname_to_card_set = {}
        self.id_to_printing = {}

        # Card Sets Index
        self.card_sets = None

        # Printing Indexes
        self.set_code_to_printings = None
        self.card_name_to_printings = None
        self.set_name_num_mv_to_printings = None

        if mtg_json_data is not None:
            self.load_mtg_json(mtg_json_data, include_online_only)
            self.rebuild_indexes()
            self.sort_indexes()

    def load_mtg_json(self, mtg_json_data, include_online_only=False):
        """Update the collection with data from mtg_json."""
        for set_data in mtg_json_data.values():
            card_set = models.CardSet(self, set_data)
            if not include_online_only and card_set.online_only:
                continue

            self.code_to_card_set[card_set.code] = card_set
            self.setname_to_card_set[card_set.name] = card_set

            for card_data in set_data['cards']:
                card = models.Card(self, card_data)
                self.name_to_card[card.name] = card

                printing = models.CardPrinting(self, card_set.code, card_data)
                self.id_to_printing[printing.id_] = printing

    def rebuild_indexes(self):
        """Rebuild the printing indexes."""
        self.card_sets = list(self.code_to_card_set.values())

        self.set_code_to_printings = collections.defaultdict(list)
        self.card_name_to_printings = collections.defaultdict(list)
        self.set_name_num_mv_to_printings = collections.defaultdict(list)

        for printing in self.id_to_printing.values():
            self.set_code_to_printings[printing.set_code].append(printing)
            self.card_name_to_printings[printing.card_name].append(printing)
            snnm_index_keys = {
                (printing.set_code, printing.card_name, printing.set_number,
                 printing.multiverseid),
                (printing.set_code, printing.card_name, None,
                 printing.multiverseid),
                (printing.set_code, printing.card_name, printing.set_number,
                 None),
                (printing.set_code, printing.card_name, None, None),
            }
            for key in snnm_index_keys:
                self.set_name_num_mv_to_printings[key].append(printing)

    def sort_indexes(self):
        """Sort the indexes."""
        self.card_sets.sort(key=lambda cset: cset.release_date)

        for printings in self.set_code_to_printings.values():
            printings.sort(key=set_code_to_printings_key)

        for printings in self.card_name_to_printings.values():
            printings.sort(key=card_name_to_printing_key)



#!/usr/bin/env python3
"""Script to manage magic collection spreadsheets."""

import argparse
import datetime
import os
import shutil

import mtg_ssm

from mtg_ssm import mtgjson
from mtg_ssm import profiling
import mtg_ssm.serialization.interface as ser_interface
from mtg_ssm.mtg import collection


MTG_SSM_DATA_PATH = os.path.expanduser(os.path.join('~', '.mtg_ssm'))


def get_args(args=None):
    """Create and return application argument parser."""
    parser = argparse.ArgumentParser(
        description='Magic Collection Spreadsheet Manager')
    parser.add_argument(
        '--version', action='version', version=mtg_ssm.__version__)

    parser.add_argument(
        '--data_path', default=MTG_SSM_DATA_PATH,
        help='Path to mtg_ssm\'s data storage folder. Default={0}'.format(
            MTG_SSM_DATA_PATH))
    parser.add_argument(
        '--include_online_only', default=False, action='store_true',
        help='Include online only sets (e.g. Masters sets) in the database.')
    parser.add_argument(
        '--profile_stats', default=False, action='store_true',
        help='Output profiling statistics.')

    format_choices = ser_interface.MtgSsmSerializer.all_formats()
    parser.add_argument(
        '--format', default='auto', choices=format_choices,
        help='File format for collection, auto will guess from the '
        'file extension.')
    parser.add_argument(
        'collection', help='Sheet to update.')

    parser.add_argument(
        '--import_format', default='auto', choices=format_choices,
        help='File format for the import file, if provided.')
    parser.add_argument(
        'imports', metavar='import', nargs='*',
        help='Optional files to read additional counts from. You may also '
        'use this to convert from one format to another. Note: counts will '
        'be added to collection, not replaced.')

    return parser.parse_args(args=args)


def build_collection(data_path, include_online_only):
    """Get a collection with current mtgjson data."""
    mtgjson.fetch_mtgjson(data_path)
    print('Reading mtgjson data.')
    mtgjsondata = mtgjson.read_mtgjson(data_path)
    return collection.Collection(
        mtgjsondata, include_online_only=include_online_only)


def process_files(args):
    """Run the requested operations."""
    coll = build_collection(args.data_path, args.include_online_only)

    for import_file in args.imports:
        _, ext = os.path.splitext(import_file)
        import_serializer_class = ser_interface.MtgSsmSerializer \
            .by_extension_and_format(ext, args.import_format)
        import_serializer = import_serializer_class(coll)
        print('Importing counts from import: %s' % import_file)
        import_serializer.read_from_file(import_file)

    _, ext = os.path.splitext(args.collection)
    serializer_class = ser_interface.MtgSsmSerializer.by_extension_and_format(
        ext, args.format)
    serializer = serializer_class(coll)

    if os.path.exists(args.collection):
        print('Reading counts from existing file.')
        serializer.read_from_file(args.collection)
        backup_name = args.collection + '.bak-{:%Y%m%d_%H%M%S}'.format(
            datetime.datetime.now())
        print('Moving existing collection to backup: %s' % backup_name)
        shutil.move(args.collection, backup_name)

    print('Writing collection to file.')
    serializer.write_to_file(args.collection)


def main():
    """Process user input and run commands.."""
    args = get_args()
    with profiling.profiled(enabled=args.profile_stats):
        process_files(args)


if __name__ == '__main__':
    main()



"""Ensure that all serializers are imported to properly set up interface."""
import mtg_ssm.serialization.interface
import mtg_ssm.serialization.csv
import mtg_ssm.serialization.deckbox
import mtg_ssm.serialization.xlsx



"""Interface definition for serializers."""

import abc
import collections
from typing import Any, Dict

from mtg_ssm.mtg import collection
from mtg_ssm.mtg import models


class Error(Exception):
    """Base error for serializers."""


class InvalidExtensionOrFormat(Error):
    """Raised if an invalid extension or format is provided."""


class DeserializationError(Error):
    """Raised when there is an error reading counts from a file."""


def find_printing(coll, set_code, name, set_number, multiverseid, strict=True):
    """Attempt to find a CardPrinting from given parameters."""
    name = name or ''
    names = [name]
    if 'Ae' in name:
        names.append(name.replace('Ae', 'Æ'))
    if 'Jo' in name:
        names.append(name.replace('Jo', 'Jö'))
    snnm_keys = []
    for name_var in names:
        snnm_keys.extend([
            (set_code, name_var, set_number, multiverseid),
            (set_code, name_var, None, multiverseid),
            (set_code, name_var, set_number, None),
            (set_code, name_var, None, None),
        ])
    for snnm_key in snnm_keys:
        found_printings = coll.set_name_num_mv_to_printings.get(snnm_key, [])
        if len(found_printings) == 1 or found_printings and not strict:
            return found_printings[0]

    return None


def coerce_counts(counts_dict):
    """Given a counts dict, coerce types to match desired input."""
    if 'multiverseid' in counts_dict:
        try:
            counts_dict['multiverseid'] = int(counts_dict['multiverseid'])
        except (TypeError, ValueError):
            pass
    for counttype in models.CountTypes:
        countname = counttype.name
        if countname in counts_dict:
            try:
                counts_dict[countname] = int(counts_dict[countname])
            except (TypeError, ValueError):
                pass
    return counts_dict


class MtgSsmSerializer(metaclass=abc.ABCMeta):
    """Abstract interface for a mtg ssm serializer."""

    _format_to_serializer = None
    _extension_to_serializer = None

    def __init__(self, coll: collection.Collection):
        self.collection = coll

    @property
    @abc.abstractmethod
    def extension(self) -> str:
        """The extension this serializer is authoritative over."""

    @property
    @abc.abstractmethod
    def format(self) -> str:
        """The format name for this serializer."""

    @abc.abstractmethod
    def write_to_file(self, filename: str) -> None:
        """Write the collection to a file."""

    @abc.abstractmethod
    def read_from_file(self, filename: str) -> None:
        """Read counts from file and add them to the collection."""

    def load_counts(self, counts: Dict[Any, Any], strict: bool=True) -> None:
        """Load counts from a dict into the collection.

        Arguments:
            counts: a dict containing key 'id' mapping to mtgjson id, and
                keys matching CountTypes name and count increment. Other keys
                are ignored.
        """
        counts = coerce_counts(counts)
        printing_id = counts.get('id')
        printing = self.collection.id_to_printing.get(printing_id)
        if printing is None:
            print('id not found for printing, searching')
            printing = find_printing(
                coll=self.collection, set_code=counts.get('set'),
                name=counts.get('name'), set_number=counts.get('number'),
                multiverseid=counts.get('multiverseid'), strict=strict)
            if printing is None:
                raise DeserializationError(
                    'Could not match id to known printing from counts: %r' %
                    counts)

        for counttype in models.CountTypes:
            countname = counttype.name
            count = counts.get(countname)
            if count:
                existing = printing.counts.get(counttype, 0)
                printing.counts[counttype] = existing + count

    @classmethod
    def _register_subclasses(cls):
        """Register formats and extensions for all subclasses."""
        cls._format_to_serializer = {}
        cls._extension_to_serializer = {}
        subclasses = collections.deque(cls.__subclasses__())
        while subclasses:
            subclass = subclasses.popleft()
            if subclass.format is not None:
                cls._format_to_serializer[subclass.format] = subclass
            if subclass.extension is not None:
                cls._extension_to_serializer[subclass.extension] = subclass
            subclasses.extend(subclass.__subclasses__())

    @classmethod
    def all_formats(cls):
        """List of all valid serializer formats."""
        if cls._format_to_serializer is None:
            cls._register_subclasses()
        formats = ['auto']
        formats.extend(cls._format_to_serializer)
        return formats

    @classmethod
    def by_extension_and_format(cls, extension: str, ser_format: str):
        """Get the appropriate serialzer for a file."""
        if cls._format_to_serializer is None:
            cls._register_subclasses()
        if ser_format == 'auto':
            serializer = cls._extension_to_serializer.get(extension.lstrip('.'))
        else:
            serializer = cls._format_to_serializer.get(ser_format)

        if serializer is None:
            raise InvalidExtensionOrFormat(
                'Cannot find serializer for format: %s and extension %s' % (
                    ser_format, extension))
        return serializer



"""XLSX serializer."""

import collections
import string

import openpyxl

from mtg_ssm.mtg import models
from mtg_ssm.serialization import interface


ALL_SETS_SHEET_HEADER = [
    'code',
    'name',
    'release',
    'block',
    'type',
    'cards',
    'unique',
    'playsets',
    'count',
]

ALL_SETS_SHEET_TOTALS = (
    ['Total'] +
    [None] * 4 +
    ['=SUM({c}3:{c}65535)'.format(c=c) for c in 'FGHI']
)


def create_all_sets(sheet, collection):
    """Create all sets sheet from collection."""
    sheet.title = 'All Sets'
    sheet.append(ALL_SETS_SHEET_HEADER)
    sheet.append(ALL_SETS_SHEET_TOTALS)
    for card_set in collection.card_sets:
        row = [
            card_set.code,
            card_set.name,
            card_set.release_date,
            card_set.block,
            card_set.type_,
            len(card_set.printings),
            '=COUNTIF(\'{}\'!A:A,">0")'.format(card_set.code),
            '=COUNTIF(\'{}\'!A:A,">=4")'.format(card_set.code),
            "=SUM('{}'!A:A)".format(card_set.code),
        ]
        sheet.append(row)


def style_all_sets(sheet):
    """Apply styles to the all sets sheet."""
    sheet.freeze_panes = sheet['C3']
    col_width_hidden = [
        ('A', 6, False),
        ('B', 24, False),
        ('C', 12, True),
        ('D', 16, True),
        ('E', 12, True),
        ('F', 6, False),
        ('G', 7, False),
        ('H', 8, False),
        ('I', 7, False),
    ]
    for col, width, hidden in col_width_hidden:
        cdim = sheet.column_dimensions[col]
        cdim.width = width
        cdim.hidden = hidden


def create_haverefs(printings):
    """Create a reference to the have cells for printings in a single set."""
    card_set = printings[0].set
    rownums = [card_set.printings.index(p) + ROW_OFFSET for p in printings]
    haverefs = [
        "'{setcode}'!A{rownum}".format(setcode=card_set.code, rownum=r)
        for r in rownums]
    return '+'.join(haverefs)


def get_references(card, exclude_sets=None):
    """Get an equation for the references to a card."""
    if card.strict_basic:
        return None  # Basics are so prolific that they overwhelm Excel

    if exclude_sets is None:
        exclude_sets = set()

    set_to_prints = collections.defaultdict(list)
    for printing in card.printings:
        if printing.set not in exclude_sets:
            set_to_prints[printing.set].append(printing)

    set_to_haveref = {k: create_haverefs(v) for k, v in set_to_prints.items()}
    if not set_to_haveref:
        return None

    references = []
    for card_set in sorted(set_to_haveref, key=lambda cset: cset.release_date):
        reference = 'IF({count}>0,"{setcode}: "&{count}&", ","")'.format(
            setcode=card_set.code, count=set_to_haveref[card_set])
        references.append(reference)
    return '=' + '&'.join(references)


ALL_CARDS_SHEET_HEADER = [
    'name',
    'have',
]


def create_all_cards(sheet, collection):
    """Create all cards sheet from collection."""
    sheet.title = 'All Cards'
    sheet.append(ALL_CARDS_SHEET_HEADER)
    # Should this be done in the collection class indexes?
    # Should card_sets not be done in the collection indexes?
    cards = sorted(collection.name_to_card.values(), key=lambda c: c.name)
    for card in cards:
        row = [
            card.name,
            get_references(card),
        ]
        sheet.append(row)


def style_all_cards(sheet):
    """Apply styles to the all cards sheet."""
    sheet.freeze_panes = sheet['B2']
    col_width_hidden = [
        ('A', 24, False),
        ('B', 32, False),
    ]
    for col, width, hidden in col_width_hidden:
        cdim = sheet.column_dimensions[col]
        cdim.width = width
        cdim.hidden = hidden


SET_SHEET_HEADER = [
    'have',
    'name',
    'id',
    'multiverseid',
    'number',
    'artist',
] + [ct.name for ct in models.CountTypes] + [
    'others',
]
COUNT_COLS = [
    string.ascii_uppercase[SET_SHEET_HEADER.index(ct.name)]
    for ct in models.CountTypes]
HAVE_TMPL = '=' + '+'.join(c + '{0}' for c in COUNT_COLS)
ROW_OFFSET = 2


def create_set_sheet(sheet, card_set):
    """Populate sheet with card information from a given set."""
    sheet.title = card_set.code
    sheet.append(SET_SHEET_HEADER)
    for printing in card_set.printings:
        rownum = card_set.printings.index(printing) + ROW_OFFSET
        row = [
            HAVE_TMPL.format(rownum),
            printing.card.name,
            printing.id_,
            printing.multiverseid,
            printing.set_number,
            printing.artist,
        ]
        for counttype in models.CountTypes:
            row.append(printing.counts.get(counttype))
        row.append(get_references(printing.card, exclude_sets={card_set}))
        sheet.append(row)


def style_set_sheet(sheet):
    """Apply styles to a set sheet."""
    sheet.freeze_panes = sheet['C2']
    col_width_hidden = [
        ('A', 5, False),
        ('B', 18, False),
        ('C', 5, True),
        ('D', 12, True),
        ('E', 8, True),
        ('F', 20, True),
        ('G', 6, False),
        ('H', 6, False),
        ('I', 10, False),
    ]
    for col, width, hidden in col_width_hidden:
        cdim = sheet.column_dimensions[col]
        cdim.width = width
        cdim.hidden = hidden


def counts_from_sheet(sheet):
    """Given an xlsx set sheet, read card counts."""
    rows = iter(sheet.rows)
    header = [cell.value for cell in next(rows)]
    for row in rows:
        row_values = [cell.value for cell in row]
        yield dict(zip(header, row_values))


class MtgXlsxSerializer(interface.MtgSsmSerializer):
    """MtgSsmSerializer for reading and writing xlsx files."""

    format = 'xlsx'
    extension = 'xlsx'

    def write_to_file(self, filename: str) -> None:
        """Write the collection to an xlsx file."""
        workbook = openpyxl.Workbook()
        all_sets_sheet = workbook.create_sheet()
        create_all_sets(all_sets_sheet, self.collection)
        style_all_sets(all_sets_sheet)
        all_cards_sheet = workbook.create_sheet()
        create_all_cards(all_cards_sheet, self.collection)
        style_all_cards(all_cards_sheet)
        for card_set in self.collection.card_sets:
            set_sheet = workbook.create_sheet()
            create_set_sheet(set_sheet, card_set)
            style_set_sheet(set_sheet)
        workbook.remove_sheet(workbook['Sheet'])
        workbook.save(filename)

    def read_from_file(self, filename: str) -> None:
        """Read collection counts from xlsx file."""
        workbook = openpyxl.load_workbook(filename=filename, read_only=True)
        for sheet in workbook.worksheets:
            if sheet.title not in self.collection.code_to_card_set:
                if sheet.title in {'Sets', 'All Sets', 'All Cards'}:
                    continue
                raise interface.DeserializationError(
                    'No known set with code {}'.format(sheet.title))
            for counts in counts_from_sheet(sheet):
                self.load_counts(counts)



"""CSV serializer."""

import csv

from mtg_ssm.mtg import models
from mtg_ssm.serialization import interface

CSV_HEADER = [
    'set',
    'name',
    'number',
    'multiverseid',
    'id',
] + [ct.name for ct in models.CountTypes]


def row_from_printing(printing):
    """Given a CardPrinting, return a csv row."""
    csv_row = {
        'set': printing.set_code,
        'name': printing.card_name,
        'number': printing.set_number,
        'multiverseid': printing.multiverseid,
        'id': printing.id_,
    }
    for counttype, count in printing.counts.items():
        if count:
            csv_row[counttype.name] = count
    return csv_row


def csv_rows_from_collection(collection):
    """Generator that yields csv rows from a collection."""
    for card_set in collection.card_sets:
        for printing in card_set.printings:
            yield row_from_printing(printing)


class MtgCsvSerializer(interface.MtgSsmSerializer):
    """MtgSsmSerializer for reading and writing csv files."""

    format = 'csv'
    extension = 'csv'

    def write_to_file(self, filename: str) -> None:
        """Write the collection to a csv file."""
        with open(filename, 'w') as csv_file:
            writer = csv.DictWriter(csv_file, CSV_HEADER)
            writer.writeheader()
            for row in csv_rows_from_collection(self.collection):
                writer.writerow(row)

    def read_from_file(self, filename: str) -> None:
        """Read collection counts from a csv file."""
        with open(filename, 'r') as csv_file:
            reader = csv.DictReader(csv_file)
            for row in reader:
                self.load_counts(row)



"""Setup script for mtg_ssm."""

import sys

import setuptools


if sys.version_info < (3, 3):
    raise Exception('Python version < 3.3 are not supported.')


# Get version information without importing the package
__version__ = None
exec(open('mtg_ssm/version.py', 'r').read())  # pylint: disable=exec-used

SHORT_DESCRIPTION = (
    'A tool to manage Magic: the Gathering collection spreadsheets.')
LONG_DESCRIPTION = open('README.rst', 'r').read()

DEPENDENCIES = [l.strip() for l in open('requirements.txt', 'r')]

if sys.version_info < (3, 5):
    DEPENDENCIES.append('typing>=3.5.0.1')
if sys.version_info < (3, 4):
    DEPENDENCIES.append('enum34>=1.1.2')

CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Console',
    'Intended Audience :: End Users/Desktop',
    'License :: OSI Approved :: MIT License',
    'Operating System :: MacOS',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: OS Independent',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Topic :: Games/Entertainment',
]

setuptools.setup(
    name='mtg_ssm',
    version=__version__,
    description=SHORT_DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author='George Leslie-Waksman',
    author_email='waksman@gwax.com',
    url='https://github.com/gwax/mtg_ssm',
    packages=setuptools.find_packages(exclude=('tests*',)),
    license='MIT',
    platforms=['any'],
    keywords='mtg magic collection tracking spreadsheet',
    classifiers=CLASSIFIERS,
    install_requires=DEPENDENCIES,
    tests_require=['pytest'],
    entry_points={
        'console_scripts': [
            'mtg-ssm = mtg_ssm.manager:main',
        ],
    },
)
